//---------------------------------------------------------------------------

#pragma hdrstop

#define _EXPORTING
#include "..\miniscope\device\MiniscopeDevice.h"
#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include <assert.h>
#include "DirectSoundScope.h"
#include "FormSettings.h"
#include "PluginSettings.h"
#include <algorithm>	// needed by Utils::in_group
#include "Utils.h"
#include <string>
#include <fstream> 
#include <json/json.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)

static const struct S_DEVICE_DLL_INTERFACE dll_interface =
	{DLL_INTERFACE_MAJOR_VERSION, DLL_INTERFACE_MINOR_VERSION};

// callback ptrs
CALLBACK_LOG lpLogFn = NULL;
CALLBACK_CONNECT lpConnectFn = NULL;
CALLBACK_DATA lpDataFn = NULL;
CALLBACK_TRIGGER lpTriggerFn = NULL;


void *callbackCookie;	///< used by upper class to distinguish library instances when receiving callbacks

void __stdcall GetMiniscopeInterfaceDescription(struct S_DEVICE_DLL_INTERFACE* interf)
{
	interf->majorVersion = dll_interface.majorVersion;
	interf->minorVersion = dll_interface.minorVersion;
}

void Log(char* txt)
{
	if (lpLogFn)
		lpLogFn(callbackCookie, txt);
}

void Data(short* data, int size)
{
	if (lpDataFn)
		lpDataFn(callbackCookie, data, size);
}

void Connect(int state, char *szMsgText)
{
	if (lpConnectFn)
		lpConnectFn(callbackCookie, state, szMsgText);
}

void Trigger(void)
{
	if (lpTriggerFn)
		lpTriggerFn(callbackCookie);
}

void __stdcall SetCallbacks(void *cookie, CALLBACK_LOG lpLog, CALLBACK_CONNECT lpConnect,
	CALLBACK_DATA lpData, CALLBACK_TRIGGER lpTrigger)
{
	assert(cookie && lpLog && lpConnect && lpData && lpTrigger);
	lpLogFn = lpLog;
	lpConnectFn = lpConnect;
	lpDataFn = lpData;
	lpTriggerFn = lpTrigger;
	callbackCookie = cookie;
	lpLogFn(callbackCookie, "dsoundscope.dll loaded. Yay!");

	scope.callbackLog = Log;
	scope.callbackData = Data;
	scope.callbackConnect = Connect;
	scope.callbackTrigger = Trigger;
}

static enum E_TRIGGER_MODE eMode[] = { TRIGGER_MODE_AUTO, TRIGGER_MODE_NORMAL, TRIGGER_MODE_SINGLE, TRIGGER_MODE_CONTINUOUS };
static enum E_TRIGGER_TYPE eTrigger[] = { TRIGGER_SLOPE_UP, TRIGGER_SLOPE_DOWN };
static enum E_COUPLING_TYPE eCoupling[] = { COUPLING_DC };

/** \sa size of DirectSoundScope::output_buffer when increasing max size
*/
static unsigned int iBuf[] = { 4410, 4800, 8000, 8820, 9600, 11025, 12000, 22050, 24000, 44100, 48000, 88200, 96000, 176400, 192000, 384000 };
float fSens[] = { 1.0f/32768.0f };
float fBasicSampl = 2.26757e-5;	// 44100
static float fSampl[] = { fBasicSampl*256, fBasicSampl*128, fBasicSampl*64, fBasicSampl*32,
	fBasicSampl*16, fBasicSampl*8, fBasicSampl*4, fBasicSampl*2, fBasicSampl };

enum { CHANNEL_CNT = DirectSoundScope::CHANNEL_CNT };
bool pChannelEnable[CHANNEL_CNT];
unsigned int pSensitivity[CHANNEL_CNT];
enum E_COUPLING_TYPE eCouplingType[CHANNEL_CNT];

/** \brief Calibration data for single sensitivity range
*/
static struct S_CALIBRATION
{
	int offset;
	float gain;
} calibration[CHANNEL_CNT][sizeof(fSens)/sizeof(fSens[0])];

void __stdcall GetDeviceCapabilities(struct S_SCOPE_CAPABILITIES **caps)
{
	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path != "")
	{
		Json::Value root;   // will contains the root value after parsing.
		Json::Reader reader;

		std::ifstream ifs(path.c_str());
		std::string strConfig((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
		ifs.close();

		bool parsingSuccessful = reader.parse( strConfig, root );
		if ( parsingSuccessful )
		{
			// read plugin settings
			const Json::Value rootPluginSettings = root["PluginSettings"];
			pluginSettings.asCaptureDevice = rootPluginSettings.get("CaptureDevice", "").asString().c_str();
			pluginSettings.samplingSpeed = rootPluginSettings.get("SamplingSpeed", pluginSettings.samplingSpeed).asInt();
			//pluginSettings.channelCnt = rootPluginSettings.get("ChannelCnt", pluginSettings.channelCnt).asInt();
			pluginSettings.fSens = rootPluginSettings.get("Sensitivity", PluginSettings::fDefaultSens).asDouble();
			pluginSettings.offset = rootPluginSettings.get("Offset", PluginSettings::DEFAULT_OFFSET).asInt();
            pluginSettings.extraLogging = rootPluginSettings.get("ExtraLogging", pluginSettings.extraLogging).asBool();
			initialPluginSettings = pluginSettings;
		}
	}
	fBasicSampl = 1.0f/pluginSettings.samplingSpeed;
	fSens[0] = initialPluginSettings.fSens;

	float fTmp = fBasicSampl;
	for (int i=ARRAY_SIZE(fSampl)-1; i>=0; i--)
	{
		fSampl[i] = fTmp;
		fTmp *= 2.0f;
	}

	static struct S_SCOPE_CAPABILITIES capabilities =
	{
        CHANNEL_CNT,
		16-1,		//bits/sample; reducing to 15 bits to leave space for calibration
		sizeof(fSens)/sizeof(fSens[0]), fSens,
		sizeof(fSampl)/sizeof(fSampl[0]), fSampl,
		sizeof(eMode)/sizeof(eMode[0]), eMode,
		sizeof(eTrigger)/sizeof(eTrigger[0]), eTrigger,
		sizeof(eCoupling)/sizeof(eCoupling[0]), eCoupling,
		sizeof(iBuf)/sizeof(iBuf[0]), iBuf
	};
	*caps = &capabilities;
}

void __stdcall ShowSettings(HANDLE parent)
{
	//MessageBox(parent, "Dsoundscope.dll has no additional settings.", "Device DLL", MB_ICONINFORMATION);
	TfrmSettings* frmSettings = new TfrmSettings(NULL);
	std::vector<std::string> devices;
	scope.GetCaptureDevicesList(devices);
	frmSettings->SetDevicesList(devices);
	frmSettings->ShowModal();
	if (scope.GetDevice() != pluginSettings.asCaptureDevice)
	{
		scope.SetDevice(pluginSettings.asCaptureDevice);
		scope.Connect();
	}                   
	delete frmSettings;
}

int __stdcall SetSensitivity(int iChannelId, int iId)
{
	if (iChannelId < 0 || iChannelId >= CHANNEL_CNT)
		return -1;
	if (iId < 0 || iId >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -2;
	return scope.SetSensitivity(iChannelId, iId);
}

int __stdcall SetSamplingPeriod(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(fSampl)/sizeof(fSampl[0])))
		return -1;
	return scope.SetSampling(sizeof(fSampl)/sizeof(fSampl[0]) - iId - 1);
}

int __stdcall SetTriggerMode(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(eMode)/sizeof(eMode[0])))
		return -1;
	return scope.SetTriggerMode(eMode[iId]);
}

int __stdcall ManualTrigger(void)
{
	return scope.ManualTrigger();
}

int __stdcall SetTriggerSource(int iChannelId, int iId)
{
	if (iChannelId < 0 || iChannelId >= CHANNEL_CNT)
		return -1;
	if (iId < 0 || iId >= (int)(sizeof(eTrigger)/sizeof(enum E_TRIGGER_TYPE)))
		return -2;
	return scope.SetTriggerType(iChannelId, eTrigger[iId]);
}

int __stdcall SetTriggerLevel(int level)
{
	return scope.SetTriggerLevel(level);	
}

int __stdcall SetBufferSize(int iId)
{
	if (iId < 0 || iId >= (int)(sizeof(iBuf)/sizeof(unsigned int)))
		return -1;
	return scope.SetBufferSize(iBuf[iId]);
}


int __stdcall Connect(void)
{
	return scope.Connect();
}

int __stdcall Disconnect(void)
{
	return scope.Disconnect();
}

int __stdcall Run(bool state)
{
	return scope.Run(state);
}

static bool bSettingsReaded = false;

static int GetDefaultSettings(struct S_SCOPE_SETTINGS* settings)
{
	settings->pChannelEnable = pChannelEnable;
	settings->pSensitivity = pSensitivity;
	settings->eCouplingType = eCouplingType;

	for (int i=0; i<CHANNEL_CNT; i++)
	{
		settings->pChannelEnable[i] = true;
		settings->eCouplingType[i] = COUPLING_DC;
		settings->pSensitivity[i] = 0;
		for (unsigned int j=0; j<sizeof(calibration[0])/sizeof(calibration[0][0]); j++)
		{
			calibration[i][j].offset = 0;
			calibration[i][j].gain = 1.0;
		}
	}
	settings->iSamplingPeriod = ARRAY_SIZE(fSampl)-1;	// "fastest" range
	settings->eTriggerMode = TRIGGER_MODE_AUTO;
	settings->eTriggerType = TRIGGER_SLOPE_UP;
	settings->iTriggerLevel = 0;
	settings->iTriggerPosition = 50;
	settings->iTriggerSrcChannel = 0;
	settings->iBufferSize = PluginSettings::DEFAULT_SAMPLING_SPEED/2;

	bSettingsReaded = true;
	return 0;
}

int __stdcall GetMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
	settings->pChannelEnable = pChannelEnable;
	settings->pSensitivity = pSensitivity;
	settings->eCouplingType = eCouplingType;

	settings->iTriggerSrcChannel = 0;

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return GetDefaultSettings(settings);

	Json::Value root;   // will contains the root value after parsing.
	Json::Reader reader;

	std::ifstream ifs(path.c_str());
	std::string strConfig((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
	ifs.close();

	bool parsingSuccessful = reader.parse( strConfig, root );
	if ( !parsingSuccessful )
		return GetDefaultSettings(settings);

	const Json::Value channels = root["Channels"];		
	for (int i=0; i<CHANNEL_CNT; i++)
	{
		const Json::Value channel = channels[i];
		settings->pSensitivity[i] = channel.get("Sensitivity", 0).asInt();
		if (settings->pSensitivity[i] >= sizeof(fSens)/sizeof(fSens[0]))
			settings->pSensitivity[i] = 0;
		settings->pChannelEnable[i] = channel.get("Enabled", true).asBool();
		int coupling = channel.get("CouplingType", eCoupling[0]).asInt();
		if (Utils::in_group(coupling, eCoupling))
			settings->eCouplingType[i] = (E_COUPLING_TYPE)coupling;
		else
			settings->eCouplingType[i] = eCoupling[0];

		// read calibration data also
		const Json::Value rootcal = channel["Calibration"];
		for (unsigned int j=0; j<sizeof(calibration[0])/sizeof(calibration[0][0]); j++)
		{
			calibration[i][j].offset = rootcal[j].get("Offset", 0).asInt();
			calibration[i][j].gain = rootcal[j].get("Gain", 1.0).asDouble();
		}
	}

	settings->iSamplingPeriod = root.get("SamplingPeriod", 0).asInt();
	if (settings->iSamplingPeriod >= sizeof(fSampl)/sizeof(fSampl[0]))
		settings->iSamplingPeriod = 0;

	int mode = root.get("TriggerMode", TRIGGER_MODE_AUTO).asInt();
	if (Utils::in_group(mode, eMode))
		settings->eTriggerMode = (E_TRIGGER_MODE)mode;
	else
		settings->eTriggerMode = TRIGGER_MODE_AUTO;

	int trigger = root.get("TriggerType", TRIGGER_SLOPE_UP).asInt();
	if (Utils::in_group(trigger, eTrigger))
		settings->eTriggerType = (E_TRIGGER_TYPE)trigger;
	else
		settings->eTriggerType = TRIGGER_SLOPE_UP;

	settings->iTriggerLevel = root.get("TriggerLevel", 0).asInt();
	if (settings->iTriggerLevel < -100 || settings->iTriggerLevel > 100)
		settings->iTriggerLevel = 0;

	int coupling = root.get("CouplingType", COUPLING_AC).asInt();
	if (Utils::in_group(coupling, eCoupling))
		settings->eCouplingType[0] = (E_COUPLING_TYPE)coupling;
	else
		settings->eCouplingType[0] = COUPLING_AC;

	settings->iBufferSize = root.get("BufferSize", 4096).asInt();
	if (!Utils::in_group(settings->iBufferSize, iBuf))
		settings->iBufferSize = iBuf[0];

	// read plugin settings
	const Json::Value rootPluginSettings = root["PluginSettings"];
	pluginSettings.asCaptureDevice = rootPluginSettings.get("CaptureDevice", "").asString().c_str();
	pluginSettings.samplingSpeed = rootPluginSettings.get("SamplingSpeed", pluginSettings.samplingSpeed).asInt();
	//pluginSettings.channelCnt = rootPluginSettings.get("ChannelCnt", pluginSettings.channelCnt).asInt();
    pluginSettings.fSens = rootPluginSettings.get("Sensitivity", PluginSettings::fDefaultSens).asDouble();
	pluginSettings.extraLogging = rootPluginSettings.get("ExtraLogging", pluginSettings.extraLogging).asBool();

    initialPluginSettings = pluginSettings;

	scope.SetDevice(pluginSettings.asCaptureDevice);

	bSettingsReaded = true;
	return 0;
}

int __stdcall SaveMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
    Json::Value root;
	Json::StyledWriter writer;
	for (int i=0; i<CHANNEL_CNT; i++)
	{
		Json::Value &channel = root["Channels"][i];
		channel["Sensitivity"] = settings->pSensitivity[i];
		channel["Enabled"] = settings->pChannelEnable[i];
		channel["CouplingType"] = settings->eCouplingType[i];
		// add calibration data also
		Json::Value &rootcal = channel["Calibration"];
		for (unsigned int j=0; j<sizeof(calibration)/sizeof(calibration[0]); j++)
		{
			rootcal[j]["Offset"] = calibration[i][j].offset;
			rootcal[j]["Gain"] = calibration[i][j].gain;
		}
	}
	root["SamplingPeriod"] = settings->iSamplingPeriod;
	root["TriggerMode"] = settings->eTriggerMode;
	root["TriggerType"] = settings->eTriggerType;
	root["TriggerLevel"] = settings->iTriggerLevel;
	root["BufferSize"] = settings->iBufferSize;

	// store plugin settings
	pluginSettings.asCaptureDevice = scope.GetDevice();
	Json::Value &ps = root["PluginSettings"];
	ps["CaptureDevice"] = pluginSettings.asCaptureDevice.c_str();
	ps["SamplingSpeed"] = pluginSettings.samplingSpeed;
	//ps["ChannelCnt"] = pluginSettings.channelCnt;
	ps["Sensitivity"] = pluginSettings.fSens;
	ps["Offset"] = pluginSettings.offset;
	ps["ExtraLogging"] = pluginSettings.extraLogging;

	std::string outputConfig = writer.write( root );

	std::string path = Utils::GetDllPath();
	path = Utils::ReplaceFileExtension(path, ".cfg");
	if (path == "")
		return -1;

    std::ofstream ofs(path.c_str());
    ofs << outputConfig;
	ofs.close();

	return 0;
}

int __stdcall StoreCalibration(int channel, int range, int offset, float gain)
{
	if (channel < 0 || channel >= CHANNEL_CNT)
		return -1;
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -1;
	calibration[channel][range].offset = offset;
	calibration[channel][range].gain = gain;
	return 0;
}

int __stdcall GetCalibration(int channel, int range, int* const offset, float* const gain)
{
	if (channel < 0 || channel >= CHANNEL_CNT)
		return -1;
	if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
		return -2;
	struct S_SCOPE_SETTINGS settings;
	if (bSettingsReaded == false)
	{
		int status = GetMiniscopeSettings(&settings);
		if (status)
			return status;
	}
	*offset = calibration[channel][range].offset;
	*gain = calibration[channel][range].gain;
	return 0;
}

int __stdcall EnableChannel(int iChannelId, bool state)
{
	if (iChannelId < 0 || iChannelId >= CHANNEL_CNT)
	{
		return -1;
	}
	pChannelEnable[iChannelId] = state;
	scope.EnableChannel(iChannelId, state);
	return 0;	
}

